#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['freezerbox',
 'freezerbox.cli',
 'freezerbox.group_by',
 'freezerbox.loaders',
 'freezerbox.stepwise']

package_data = \
{'': ['*']}

install_requires = \
['appcli',
 'appdirs',
 'autoprop',
 'autosnapgene',
 'biopython',
 'configurator',
 'docopt',
 'entrypoints',
 'exmemo',
 'inform~=1.21',
 'more_itertools~=8.0',
 'networkx',
 'numpy',
 'pandas',
 'parsy',
 'sklearn',
 'stepwise~=0.23',
 'tidyexc',
 'tqdm',
 'voluptuous']

extras_require = \
{'doc': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'test': ['pytest',
          'pytest-cov',
          'coveralls',
          'parametrize_from_file',
          're_assert']}

entry_points = \
{'console_scripts': ['freezerbox = freezerbox.cli.main:main'],
 'freezerbox.databases': ['excel = freezerbox.loaders.excel'],
 'freezerbox.make': ['dilute = freezerbox.stepwise.dilute:Dilute',
                     'order = freezerbox.stepwise.make:OrderMaker',
                     'sw = freezerbox.stepwise.make:StepwiseMaker'],
 'stepwise.protocols': ['protocols = freezerbox.stepwise:Plugin']}

setup(name='freezerbox',
      version='0.18.0',
      description='Database of DNA sequences.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/freezerbox',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='~=3.8',
     )
